/*-----------------09-07-94 09:36am-----------------
 * This is include file "defines" a generic singly linked list. It is to be
 * used in conjunction with the file 'slist.h', which "declares" a generic
 * singly linked list. To use it, define "TYPE" for the element you wish the
 * SList to contain, and then include this file. In order for an element to
 * be containable, it must be a struct. You must ensure that it has some
 * necessary fields, and certain functions must be declared and perform
 * important operations.
 * Before an SList can be "defined" it must be "declared" all the code
 * necessary to do this is as follows;

	typedef struct element {
		struct element *next; // next element in the list
		...other fields of "element"
	} element;

	void elementInit(element *e,...other parameters);
	void elementDone(element *e);
	void elementCopy(element *e,element *o);

	#define TYPE  element
	#include "slist.h"

 * All the necessary code to "define" an SList of 'element' is shown below. Note
 * that all fields, functions, and operations shown are essential for the SList
 * to work, except 'elementInit'. The 'elementInit' function is advisable but
 * not necessary.

	void elementInit(element *e,...other parameters)
	{
		assert(e!=NULL);
	#ifndef NDEBUG
		e->next=NULL;
	#endif//NDEBUG
		...other initialization code for "element" (init "e")
	}

	void elementDone(element *e)
	{
		assert(e!=NULL);
		...other destruction code for "element" (destroy "e")
	}

	void elementCopy(element *e,element *o)
	{
		assert(e!=NULL);
		assert(o!=NULL);
	#ifndef NDEBUG
		e->next=NULL;
	#endif//NDEBUG
		...other copying code for "element" (copy "o" to "e")
	}

	#define TYPE  element
	#include "slist.c"

 * This effectively automaticaly creates function "definitions" for the
 * following functions.

	void elementSListInit(elementSList *l);
	//Initializes a element list

	void elementSListDone(elementSList *l);
	//destroys a element list, freeing any included elements

	void elementSListCopy(elementSList *l,elementSList *o);
	//copys element list 'o' into element list 'l'

	void elementSListRem(elementSList *l,element *e);
	//removes element 'e' from the list 'l'

	void elementSListAdd(elementSList *l,element *e);
	//adds element 'e' to the start of list 'l'

	void elementSListApp(elementSList *l,element *e);
	//appends element 'e' to the end of list 'l'

	element *elementSListPop(elementSList *l);
	//returns and pops a element from the top of the list 'l'

	element *elementSListPul(elementSList *l);
	//returns and puls a element from the bottom of the list 'l'

	void elementSListAft(elementSList *l,element *p,element *e);
	//inserts element 'e' after 'p' in list 'l'.
	//If 'p'=NULL, 'e'is put at the front of the list

	void elementSListBef(elementSList *l,element *p,element *e);
	//inserts element 'e' before 'p' in list 'l'.
	//If 'p'=NULL, 'e'is put at the end of the list

	TYPE *elementSListPrev(elementSList *l,element *e);
	//returns the element before 'e' in the list 'l'.
	//If 'e'=NULL, returns the end of the list

 * --------------------------------------------------*/

/*-----------------09-07-94 09:36am-----------------
 * Added features: Turning off unused functions.
 *
 * Every function, apart from SListInit and SListDone can now be turned
 * off to reduce code size created by unused functions. This is done
 * by defining a name "SLISTxxx_OFF" for each feature disabled. Note that
 * these constants are are undef'ed at the end of this file, so if you
 * have multiple SList instances in the one file, you must ensure all the
 * required disabling names are defined before this file is included.
 * Note that SListPrev cannot be turned off as it is needed by other routines.
 * Here is an example:

	#define SLISTCOPY_OFF
	#define SLISTAPP_OFF
	#define TYPE  element
	#define FUNC(n) element##n
	#include "slist.h"

	#define SLISTCOPY_OFF
	#define SLISTAPP_OFF
	#define TYPE  element
	#define FUNC(n) element##n
	#include "slist.c"

 * This declares and defines an elementSList as above, but without any
 * elementSListCopy or elementSListApp functions. Note that the disabling
 * #defines are required before both including slist.h AND including
 * slist.c
 * --------------------------------------------------*/

/*-----------------01/02/96 17:02-------------------
 * Added features: the FUNC(n) definition is no longer required.
 *
 * I now know how to trick the pre-processor to do the job by using
 * "double indirection" with JOIN and JOIN2 (see below).
 * --------------------------------------------------*/


#ifndef TYPE
#error TYPE needs to be defined
#endif

#define FUNC(f)   JOIN(TYPE,f)
#define JOIN(x,y) JOIN2(x,y)
#define JOIN2(x,y) x##y

#define TYPESList FUNC(SList)

#include <assert.h>
#include <stdlib.h>


void FUNC(SListInit)(TYPESList *l)
{
	assert(l!=NULL);
	l->next=NULL;
}


void FUNC(SListDone)(TYPESList *l)
{
	TYPE *e;

	e=l->next;
	while (e!=NULL) {
		l->next=e->next;
		FUNC(Done)(e);
		free(e);
		e=l->next;
	}
#ifndef NDEBUG
	l->next=NULL;
#endif//NDEBUG
}


#ifndef SLISTCOPY_OFF
void FUNC(SListCopy)(TYPESList *l,TYPESList *o)
{
	TYPE *e,*n,*prev;

	assert(o!=NULL);
	FUNC(SListInit)(l);
	e=o->next;
	if (e!=NULL) {
		n=malloc(sizeof(*n));
		FUNC(Copy)(n,e);
		l->next=prev=n;
		e=e->next;
		while (e!=NULL) {
			n=malloc(sizeof(*n));
			FUNC(Copy)(n,e);
			prev=prev->next=n;
			e=e->next;
		}
		n->next=NULL;
	}
}
#endif //SLISTCOPY_OFF


#ifndef SLISTREM_OFF
void FUNC(SListRem)(TYPESList *l,TYPE *e)
{
	TYPE *prev;

	assert(e!=NULL);
	prev=FUNC(SListPrev)(l,e);
	if (prev!=NULL) {
		prev->next=e->next;
	} else {
		l->next=e->next;
	}
#ifndef NDEBUG
	e->next=NULL;
#endif//NDEBUG
}
#endif //SLISTREM_OFF


#ifndef SLISTADD_OFF
void FUNC(SListAdd)(TYPESList *l,TYPE *e)
{
	assert(e!=NULL);
	e->next=l->next;
	l->next=e;
}
#endif //SLISTADD_OFF


#ifndef SLISTAPP_OFF
void FUNC(SListApp)(TYPESList *l,TYPE *e)
{
	TYPE *prev;

	assert(e!=NULL);
	prev=FUNC(SListPrev)(l,NULL);
	e->next=NULL;
	if (prev!=NULL) {
		prev->next=e;
	} else {
		l->next=e;
	}
}
#endif //SLISTAPP_OFF


#ifndef SLISTPOP_OFF
TYPE *FUNC(SListPop)(TYPESList *l)
{
	TYPE *e;

	e=l->next;
	if (e!=NULL) {
		l->next=e->next;
#ifndef NDEBUG
		e->next=NULL;
#endif//NDEBUG
	}
	return(e);
}
#endif //SLISTPOP_OFF


#ifndef SLISTPUL_OFF
TYPE *FUNC(SListPul)(TYPESList *l)
{
	TYPE *e,*prev;

	e=FUNC(SListPrev)(l,NULL);
	prev=FUNC(SListPrev)(l,e);
	if (e!=NULL) {
		if (prev!=NULL) {
			prev->next=NULL;
		} else {
			l->next=NULL;
		}
#ifndef NDEBUG
		e->next=NULL;
#endif//NDEBUG
	}
	return(e);
}
#endif //SLISTPUL_OFF


#ifndef SLISTAFT_OFF
void FUNC(SListAft)(TYPESList *l,TYPE *p,TYPE *e)
{
	assert(e!=NULL);
	if (p!=NULL) {
		e->next=p->next;
		p->next=e;
	} else {
		e->next=l->next;
		l->next=e;
	}
}
#endif //SLISTAFT_OFF


#ifndef SLISTBEF_OFF
void FUNC(SListBef)(TYPESList *l,TYPE *p,TYPE *e)
{
	TYPE *prev;

	assert(e!=NULL);
	prev=FUNC(SListPrev)(l,p);
	e->next=p;
	if (prev!=NULL) {
		prev->next=e;
	} else {
		l->next=e;
	}
}
#endif //SLISTBEF_OFF


TYPE *FUNC(SListPrev)(TYPESList *l,TYPE *e)
{
	TYPE *prev;

	prev=l->next;
	if (e!=prev) {
		while (prev->next!=e) {
			assert(prev!=NULL); //end of list reached and e not found
			prev=prev->next;
		};
		return(prev);
	} else {
		return(NULL);
	}
}


#undef SLISTCOPY_OFF
#undef SLISTREM_OFF
#undef SLISTADD_OFF
#undef SLISTAPP_OFF
#undef SLISTPOP_OFF
#undef SLISTPUL_OFF
#undef SLISTAFT_OFF
#undef SLISTBEF_OFF

#undef TYPESList
#undef FUNC
#undef TYPE
